/*  $Header$
 *
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/version.h>
#ifdef MODULE
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/kernel.h>
#include <linux/if_arp.h>
#include <asm/uaccess.h>

#include <isl_ioctl.h>
#include <isl_gen.h>
#include <islpci_mgt.h>
#include <bloboid.h>		// additional types and defs for isl38xx fw


#ifdef WIRELESS_IOCTLS

#undef IW_MAX_BITRATES
#define	IW_MAX_BITRATES	16

#define MAX_KEY_SIZE	16
#define	MAX_KEYS	4
char key_table[MAX_KEYS][MAX_KEY_SIZE];
char nick_name[32];
int key_on = 0;
int key_open = 1;
int key_current = 0;
int key_size[MAX_KEYS];

/******************************************************************************
        Global variable definition section
******************************************************************************/
extern int pc_debug;

struct iw_priv_args privtab[] = {
    // cmd                   set_args               get_args                     name
    {SIOCIWFIRSTPRIV + 0x00, IW_PRIV_TYPE_INT | 1, 0, "setauthenten"},	// 0x8BE0
    {SIOCIWFIRSTPRIV + 0x01, 0, IW_PRIV_TYPE_INT | 1, "getauthenten"},	// 0x8BE1
    {SIOCIWFIRSTPRIV + 0x02, IW_PRIV_TYPE_INT | 1, 0, "setunencrypt"},	// 0x8BE2
    {SIOCIWFIRSTPRIV + 0x03, 0, IW_PRIV_TYPE_INT | 1, "getunencrypt"},
    {SIOCIWFIRSTPRIV + 0x04, IW_PRIV_TYPE_INT | 1, 0, "setprivinvok"},
    {SIOCIWFIRSTPRIV + 0x05, 0, IW_PRIV_TYPE_INT | 1, "getprivinvok"},
    {SIOCIWFIRSTPRIV + 0x06, IW_PRIV_TYPE_INT | 1, 0, "setdefkeyid"},
    {SIOCIWFIRSTPRIV + 0x07, 0, IW_PRIV_TYPE_INT | 1, "getdefkeyid"},
    {SIOCIWFIRSTPRIV + 0x08, IW_PRIV_TYPE_CHAR | 16, 0, "setdefkeyx"},
    {SIOCIWFIRSTPRIV + 0x09, IW_PRIV_TYPE_INT | 1, IW_PRIV_TYPE_CHAR | 16,
     "getdefkeyx"},
    {SIOCIWFIRSTPRIV + 0x0A, 0, IW_PRIV_TYPE_INT | 4, "getprivstat"},
    {SIOCIWFIRSTPRIV + 0x0B, 0, IW_PRIV_TYPE_BYTE | IW_MAX_BITRATES,
     "getsuprates"},
    {SIOCIWFIRSTPRIV + 0x0C, IW_PRIV_TYPE_INT | 1, 0, "setdtimper"},
    {SIOCIWFIRSTPRIV + 0x0D, 0, IW_PRIV_TYPE_INT | 1, "getdtimper"},
    {SIOCIWFIRSTPRIV + 0x0E, IW_PRIV_TYPE_INT | 1, 0, "setbeaconper"},
    {SIOCIWFIRSTPRIV + 0x0F, 0, IW_PRIV_TYPE_INT | 1, "getbeaconper"},
    {SIOCIWFIRSTPRIV + 0x10, IW_PRIV_TYPE_CHAR | 12, 0, "wdslinkadd"},
    {SIOCIWFIRSTPRIV + 0x11, IW_PRIV_TYPE_CHAR | 12, 0, "wdslinkdel"},
    {SIOCIWFIRSTPRIV + 0x12, IW_PRIV_TYPE_CHAR | 12, 0, "eapauthen"},
    {SIOCIWFIRSTPRIV + 0x13, IW_PRIV_TYPE_CHAR | 12, 0, "eapunauth"},
    {SIOCIWFIRSTPRIV + 0x14, IW_PRIV_TYPE_INT | 1, 0, "setdot1xen"},
    {SIOCIWFIRSTPRIV + 0x15, 0, IW_PRIV_TYPE_INT | 1, "getdot1xen"},
    {SIOCIWFIRSTPRIV + 0x16, IW_PRIV_TYPE_CHAR | 32, 0, "setstakeyx"},	// <address>:<key_nr>:<key>
    {SIOCIWFIRSTPRIV + 0x17, IW_PRIV_TYPE_CHAR | 32, IW_PRIV_TYPE_CHAR | 32, "getstakeyx"},	// <address>:<key_nr> -> <key>
    {SIOCIWFIRSTPRIV + 0x18, IW_PRIV_TYPE_INT | 1, 0, "setfixedrate"},
    {SIOCIWFIRSTPRIV + 0x19, 0, IW_PRIV_TYPE_INT | 1, "getfixedrate"},
/*
        { SIOCIWFIRSTPRIV + 0x18, IW_PRIV_TYPE_INT | 1, 0, "setfragthresh" },
        { SIOCIWFIRSTPRIV + 0x19, 0, IW_PRIV_TYPE_INT | 1, "getfragthresh" },
*/
    {SIOCIWFIRSTPRIV + 0x1A, IW_PRIV_TYPE_INT | 1, 0, "setpsm"},
    {SIOCIWFIRSTPRIV + 0x1B, 0, IW_PRIV_TYPE_INT | 1, "getpsm"},
    {SIOCIWFIRSTPRIV + 0x1C, IW_PRIV_TYPE_INT | 1, 0, "setnonerp"},
    {SIOCIWFIRSTPRIV + 0x1D, 0, IW_PRIV_TYPE_INT | 1, "getnonerp"},
    {SIOCIWFIRSTPRIV + 0x1E, 0, IW_PRIV_TYPE_BYTE | 4, "getcountrystr"},
//      { SIOCIWFIRSTPRIV + 0x1F, 0, IW_PRIV_TYPE_BYTE | 16, "checktraps" },
    {SIOCIWFIRSTPRIV + 0x1F, IW_PRIV_TYPE_INT | 1, 0, "setpc_debug"},
//      { SIOCIWFIRSTPRIV + 0x1F, 0, IW_PRIV_TYPE_INT | 1, "setpc_debug" }
};

/******************************************************************************
    Driver Management IOCTL functions
******************************************************************************/
#ifdef WIRELESS_EXT

struct iw_statistics *islpci_get_wireless_stats(struct net_device *nwdev)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;

    char eth_addr[6];		/* Access point address */
    int operation, error;
    struct obj_bss bss, *pBss = NULL;
    unsigned long dlen, noiseFloor, *pNoiseFloor;
    unsigned long flags;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_FUNCTION_CALLS, "%s ->islpci_get_wireless_stats\n",
	  nwdev->name);
#endif

    // lock the driver code
    driver_lock(&private_config->slock, &flags);

    error = isl_ioctl_getbssid(nwdev, eth_addr);

    if (error) {
	printk("%s: stats Error ioctl_getbssid\n", nwdev->name);
    } else {
	memset(&private_config->wirelesss_stats, 0,
	       sizeof(private_config->wirelesss_stats));

	memcpy(bss.address, eth_addr, 6);

	// send a set object to set the bsstype by placing it in the transmit queue
	islpci_mgt_queue(private_config, PIMFOR_OP_GET, DOT11_OID_BSSFIND,
			 0, &bss, sizeof(struct obj_bss), 0);

	// enter sleep mode after supplying oid and local buffers for reception
	// to the management queue receive mechanism
	if (error = islpci_mgt_response(private_config, DOT11_OID_BSSFIND,
					&operation, (void *) &pBss, &dlen,
					&entry), error != 0) {
	    // error in getting the response, return the value
	    printk("get_wireless_stats error response\n");
	    islpci_mgt_release(private_config, entry);
	    driver_unlock(&private_config->slock, &flags);
	    return &private_config->wirelesss_stats;
	} else if (operation == PIMFOR_OP_ERROR) {
	    printk("get_wireless_stats PIMFOR_OP_ERROR\n");
	    islpci_mgt_release(private_config, entry);
	    driver_unlock(&private_config->slock, &flags);
	    return &private_config->wirelesss_stats;
	}
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING,
	      "%s: channel: %d , ssid: %.*s , Qual: %d, Rssi: %d\n",
	      nwdev->name, pBss->channel, pBss->ssid[0],
	      (char *) &pBss->ssid[1], (int) pBss->quality,
	      (int) pBss->rssi);
	DEBUG(SHOW_TRACING,
	      "%s: beacon_period: %d , dtim_period: %d, capinfo: 0x%02x, rates: 0x%02x\n",
	      nwdev->name, pBss->beacon_period, pBss->dtim_period,
	      pBss->capinfo, pBss->rates);
#endif

	private_config->wirelesss_stats.qual.qual = pBss->quality;
	private_config->wirelesss_stats.qual.level = pBss->rssi;
	islpci_mgt_release(private_config, entry);

	// send a set object to set the bsstype by placing it in the transmit queue
	noiseFloor = 0;
	islpci_mgt_queue(private_config, PIMFOR_OP_GET,
			 DOT11_OID_NOISEFLOOR, 0, &noiseFloor,
			 sizeof(noiseFloor), 0);

	// enter sleep mode after supplying oid and local buffers for reception
	// to the management queue receive mechanism
	if (error =
	    islpci_mgt_response(private_config, DOT11_OID_NOISEFLOOR,
				&operation, (void *) &pNoiseFloor, &dlen,
				&entry), error != 0) {
	    // error in getting the response, return the value
	    printk("%s, get_wireless_stats error response\n", nwdev->name);
	    islpci_mgt_release(private_config, entry);
	    driver_unlock(&private_config->slock, &flags);
	    return &private_config->wirelesss_stats;
	} else if (operation == PIMFOR_OP_ERROR) {
	    printk("%s, get_wireless_stats PIMFOR_OP_ERROR\n",
		   nwdev->name);
	    islpci_mgt_release(private_config, entry);
	    driver_unlock(&private_config->slock, &flags);
	    return &private_config->wirelesss_stats;
	}
//      *pNoiseFloor = 0x88;      
	private_config->wirelesss_stats.qual.noise =
	    (unsigned char) *pNoiseFloor;
	private_config->wirelesss_stats.qual.updated = 0x7;

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s, NoiseFloor: %lu, dlen = %lu\n",
	      nwdev->name,
	      (unsigned long) private_config->wirelesss_stats.qual.noise,
	      dlen);
#endif

    }

    private_config->wirelesss_stats.status = 1;
    private_config->wirelesss_stats.discard.nwid = 0L;
    private_config->wirelesss_stats.discard.code = 0L;
    private_config->wirelesss_stats.discard.misc = 0L;

    islpci_mgt_release(private_config, entry);

    // unlock the driver code
    driver_unlock(&private_config->slock, &flags);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_FUNCTION_CALLS, "%s: <-islpci_get_wireless_stats\n",
	  nwdev->name);
#endif

    return &private_config->wirelesss_stats;
}
#endif				/* WIRELESS_EXT */


static int
ioctl_set_encode(struct net_device *nwdev, struct iw_point *erq, char *key)
{
    struct iwreq wrq;
    int error;
    int index = (erq->flags & IW_ENCODE_INDEX) - 1;

    if (erq->length > 0) {
	/* Check the size of the key */
	if (erq->length > MAX_KEY_SIZE)
	    return (-EINVAL);

	/* Check the index */
	if ((index < 0) || (index >= MAX_KEYS))
	    index = key_current;

	/* Copy the key in the driver */
	memcpy(key_table[index], key, erq->length);
	key_table[index][erq->length] = 0;
	key_size[index] = erq->length;
	key_on = 1;
    } else {
	/* Do we want to just set the current key ? */
	if ((index >= 0) && (index < MAX_KEYS)) {
	    if (key_size[index] > 0) {
		key_current = index;
		key_on = 1;
	    } else
		return (-EINVAL);
	}
    }

    /* Read the flags */
    if (erq->flags & IW_ENCODE_DISABLED)
	key_on = 0;		/* disable encryption */
    if (erq->flags & IW_ENCODE_RESTRICTED)
	key_open = 0;		/* disable open mode */
    if (erq->flags & IW_ENCODE_OPEN)
	key_open = 1;		/* enable open mode */

    printk
	("ioctl_set_encode : key_on=%d, key_open= %d, KeyId: %d, Key= %.*s\n",
	 key_on, key_open, key_current, key_size[key_current],
	 &key_table[key_current][0]);

    wrq.u.data.pointer = &key_table[key_current][0];
    error = isl_ioctl_setdefkeyx(nwdev, &wrq, key_current);
    if (error)
	return error;

    wrq.u.data.pointer = (void *) &key_on;
    return (isl_ioctl_setprivinvoke(nwdev, &wrq));
}

static int
ioctl_get_encode(struct net_device *dev, struct iw_point *erq, char *key)
{
    int index = (erq->flags & IW_ENCODE_INDEX) - 1;

    /* Set the flags */
    erq->flags = 0;
    if (key_on == 0)
	erq->flags |= IW_ENCODE_DISABLED;
    if (key_open == 0)
	erq->flags |= IW_ENCODE_RESTRICTED;
    else
	erq->flags |= IW_ENCODE_OPEN;

    /* Which key do we want */
    if ((index < 0) || (index >= MAX_KEYS))
	index = key_current;
    erq->flags |= index + 1;

    /* Copy the key to the user buffer */
    erq->length = key_size[index];
    memcpy(key, key_table[index], key_size[index]);

    return (0);
}

int isl_ioctl_getrange(struct net_device *nwdev, struct iw_range *range)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    struct iw_range range_obj;
    char *data;
    unsigned long dlen;
    int rvalue, counter, operation;

    // request the device for the rates
    islpci_mgt_queue(private_config, PIMFOR_OP_GET, DOT11_OID_RATES, 0,
		     NULL, IW_MAX_BITRATES + 1, 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    if (rvalue = islpci_mgt_response(private_config, DOT11_OID_RATES,
				     &operation, (void *) &data, &dlen,
				     &entry), rvalue != 0) {
	// error in getting the response, return the value
	printk("Error mgt_response getrange\n");
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	printk("Error PIMFOR_OP_ERROR getrange\n");
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "getrange: bufaddr %p \n", range);
#endif

    // set the wireless extension version number
    range_obj.we_version_source = SUPPORTED_WIRELESS_EXT;
    range_obj.we_version_compiled = WIRELESS_EXT;

    /* Need to get better values */
    range_obj.max_qual.qual = 1;
    range_obj.max_qual.level = 0;
    range_obj.max_qual.noise = 0;

    range_obj.avg_qual.qual = 1;
    range_obj.avg_qual.level = 50;
    range_obj.avg_qual.noise = 101;


    for (counter = 0; counter < IW_MAX_BITRATES; counter++, data++) {
	// check whether the rate value is set
	if (*data) {
#if VERBOSE > SHOW_ERROR_MESSAGES
	    DEBUG(SHOW_TRACING, "rate: basic %i value %x \n",
		  (*data & 0x80) >> 7, *data & 0x7F);
#endif

	    // convert the bitrates to Mbps
	    range_obj.bitrate[counter] = (double) (*data & 0x7F);
	    range_obj.bitrate[counter] *= 1000000;

	    // increment the number of rates counter
	    range_obj.num_bitrates++;
	}
    }

    // return the data if the destination buffer is defined
    if (copy_to_user((void *) range, (void *) &range_obj,
		     sizeof(struct iw_range))) {
	printk("Error copy getrange\n");
	islpci_mgt_release(private_config, entry);
	return -EFAULT;
    }


    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_setrange(struct net_device *nwdev, struct iw_range *range)
{
    return 0;
}

#ifdef WIRELESS_EXT
int
isl_ioctl_setstats(struct net_device *nwdev, struct iw_statistics *stats)
{
    return 0;
}
#endif

int isl_ioctl_getauthenten(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "-> Get Authenticate\n");
#endif

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_AUTHENABLE, 0, NULL, 4, 0);
    if (rvalue = islpci_mgt_response(private_config, DOT11_OID_AUTHENABLE,
				     &operation, (void *) &data, &dlen,
				     &entry), rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    *flag = (long) *data;
    wrq->u.data.length = 1;
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get Authenticate enable: %i\n", (int) *flag);
#endif

    islpci_mgt_release(private_config, entry);
    DEBUG(SHOW_TRACING, "<- Get Authenticate\n");
    return 0;
}

int isl_ioctl_setauthenten(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    char filter[4] = { 0, 0, 0, 0 };
    int rvalue, operation;
    void *data;
    unsigned long dlen;

    filter[0] = (char) *flag;
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set Authenticate enable: %i\n", (int) filter[0]);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET, DOT11_OID_AUTHENABLE,
		     0, &filter[0], 4, 0);
    rvalue = islpci_mgt_response(private_config, DOT11_OID_AUTHENABLE,
				 &operation, (void *) &data, &dlen,
				 &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_getprivfilter(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_EXUNENCRYPTED, 0, NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_EXUNENCRYPTED,
			    &operation, (void *) &data, &dlen, &entry),
	rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    *flag = (long) *data;
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get Privacy filter: %i\n", (int) *flag);
#endif

    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_setprivfilter(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    char filter[4] = { 0, 0, 0, 0 };
    int rvalue, operation;
    void *data;
    unsigned long dlen;

    filter[0] = (char) *flag;
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set Privacy filter: %i\n", (int) filter[0]);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_EXUNENCRYPTED, 0, &filter[0], 4, 0);
    rvalue =
	islpci_mgt_response(private_config, DOT11_OID_EXUNENCRYPTED,
			    &operation, (void *) &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_getprivinvoke(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_PRIVACYINVOKED, 0, NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_PRIVACYINVOKED,
			    &operation, (void *) &data, &dlen, &entry),
	rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    *flag = (long) *data;
    wrq->u.data.length = 1;
    islpci_mgt_release(private_config, entry);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get Privacy invoked: %i\n", (int) *flag);
#endif
    return 0;
}

int isl_ioctl_setprivinvoke(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    char invoke[4] = { 0, 0, 0, 0 };
    int rvalue, operation;
    void *data;
    unsigned long dlen;

    invoke[0] = (char) *flag;
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set Privacy Key invoked: %i\n", (int) invoke[0]);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_PRIVACYINVOKED, 0, &invoke[0], 4, 0);
    rvalue =
	islpci_mgt_response(private_config, DOT11_OID_PRIVACYINVOKED,
			    &operation, (void *) &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;

    wrq->u.data.length = 1;
    return rvalue;
}

int isl_ioctl_getdefkeyid(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *key = (long *) wrq->u.data.pointer;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET, DOT11_OID_DEFKEYID, 0,
		     NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_DEFKEYID, &operation,
			    (void *) &data, &dlen, &entry), rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    *key = (long) *data;
    wrq->u.data.length = 1;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get Default Key ID %i\n", (int) *key);
#endif

    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_setdefkeyid(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *key = (long *) wrq->u.data.pointer;
    char keyid[4] = { 0, 0, 0, 0 };
    int rvalue, operation;
    void *data;
    unsigned long dlen;

    keyid[0] = (char) *key;
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set Default Key ID: %i\n", (int) keyid[0]);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET, DOT11_OID_DEFKEYID, 0,
		     &keyid[0], 4, 0);
    rvalue =
	islpci_mgt_response(private_config, DOT11_OID_DEFKEYID, &operation,
			    &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int
isl_ioctl_getdefkeyx(struct net_device *nwdev, struct iwreq *wrq,
		     int keyId)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    int *key = (int *) wrq->u.data.pointer;
    int rvalue, operation;
    struct obj_key *pkey;
    unsigned long oid;
    unsigned long dlen;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get Default Key %i\n", (int) *key);
#endif

    switch (*key) {
    case 0:
	oid = DOT11_OID_DEFKEY1;
	break;

    case 1:
	oid = DOT11_OID_DEFKEY2;
	break;

    case 2:
	oid = DOT11_OID_DEFKEY3;
	break;

    case 3:
	oid = DOT11_OID_DEFKEY4;
	break;

    default:
	return -EOPNOTSUPP;
    }

    // store the key value for future write operations
    key_current = *key;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET, oid, 0, NULL,
		     sizeof(struct obj_key), 0);
    if (rvalue =
	islpci_mgt_response(private_config, oid, &operation,
			    (void *) &pkey, &dlen, &entry), rvalue != 0) {
	// an error occurred, return the value
	printk("Error Get Default Key\n");
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	printk("PIMFOR_OP_ERROR Get Default Key\n");
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    if (copy_to_user((char *) key, (void *) pkey->key, pkey->length)) {
	printk("Error copy getdefkeyx\n");
	islpci_mgt_release(private_config, entry);
	return -EFAULT;
    }

    wrq->u.data.length = sizeof(pkey->key);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Encryption Key Nr. %d :  %s\n", key_current,
	  pkey->key);
#endif

    islpci_mgt_release(private_config, entry);
    return 0;
}

int
isl_ioctl_setdefkeyx(struct net_device *nwdev, struct iwreq *wrq,
		     int keyId)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    char *key = (char *) wrq->u.data.pointer;
    void *data;
    unsigned long dlen;
    struct obj_key defkey;
    unsigned long oid;
    int rvalue, operation;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "->Set Default Key %i \n", keyId);
#endif
    printk("->Set Default Key %i \n", keyId);

    switch (keyId) {
    case 0:
	oid = DOT11_OID_DEFKEY1;
	break;

    case 1:
	oid = DOT11_OID_DEFKEY2;
	break;

    case 2:
	oid = DOT11_OID_DEFKEY3;
	break;

    case 3:
	oid = DOT11_OID_DEFKEY4;
	break;

    default:
	return -EOPNOTSUPP;
    }

    defkey.type = DOT11_PRIV_WEP;
    defkey.length =
	strlen(key) >
	sizeof(defkey.key) ? sizeof(defkey.key) : strlen(key);

#if 0
    if (copy_from_user(defkey.key, key, defkey.length))	// to, from , length
    {
	printk("isl_ioctl_setdefkeyx: Error copy_from_user\n");
	return -EFAULT;
    }
#else
    memcpy(defkey.key, key, defkey.length);	// to, from ,size
#endif

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "isl_ioctl_setdefkeyx: key: %s, Keylength: %d\n",
	  defkey.key, defkey.length);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET, oid, 0, &defkey,
		     sizeof(struct obj_key), 0);

    rvalue =
	islpci_mgt_response(private_config, oid, &operation, &data, &dlen,
			    &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR) {
	printk("isl_ioctl_setdefkeyx: PIMFOR_OP_ERROR\n");
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "%s: <-setdefkeyx : Key-Nr.: %d, %s\n",
	  __FUNCTION__, keyId, key);
#endif

    return rvalue;
}

int isl_ioctl_getprivstat(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    int *stat = (int *) wrq->u.data.pointer;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "-> getprivstat\n");
#endif

    // send a get object to get transmit rejected frame count
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_PRIVTXREJECTED, 0, NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_PRIVTXREJECTED,
			    &operation, (void *) &data, &dlen, &entry),
	rvalue != 0) {
	printk("Error mgt_response getprivstat\n");
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	printk("Error PIMFOR_OP_ERROR getprivstat\n");
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "MPDU Tx Rejected %i\n", (int) *data);
#endif
    *stat = (int) *data;
    stat++;
    islpci_mgt_release(private_config, entry);

    // send a get object to get rejected unencrypted frame count
    islpci_mgt_queue(private_config, PIMFOR_OP_GET, DOT11_OID_PRIVRXPLAIN,
		     0, NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_PRIVRXPLAIN,
			    &operation, (void *) &data, &dlen, &entry),
	rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "MPDU Rx Plain %i\n", (int) *data);
#endif
    *stat = (int) *data;
    stat++;
    islpci_mgt_release(private_config, entry);

    // send a get object to get rejected crypted frame count
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_PRIVRXFAILED, 0, NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_PRIVRXFAILED,
			    &operation, (void *) &data, &dlen, &entry),
	rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "MPDU Rx Failed %i\n", (int) *data);
#endif
    *stat = (int) *data;
    stat++;
    islpci_mgt_release(private_config, entry);

    // send a get object to get received no key frames
    islpci_mgt_queue(private_config, PIMFOR_OP_GET, DOT11_OID_PRIVRXNOKEY,
		     0, NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_PRIVRXNOKEY,
			    &operation, (void *) &data, &dlen, &entry),
	rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	printk("Error PIMFOR_OP_ERROR getrange\n");
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "MPDU Rx No Key %i\n", (int) *data);
#endif
    *stat = (int) *data;
    islpci_mgt_release(private_config, entry);
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "<- getprivstat\n");
#endif

    return 0;
}

int isl_ioctl_getsuprates(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    char *rates = (char *) wrq->u.data.pointer;
    char *data;
    unsigned long dlen;
    int rvalue, operation;

    // request the device for the rates
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_SUPPORTEDRATES, 0, NULL,
		     IW_MAX_BITRATES + 1, 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_SUPPORTEDRATES,
			    &operation, (void *) &data, &dlen, &entry),
	rvalue != 0) {
	// error in getting the response, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    if (copy_to_user(rates, data, IW_MAX_BITRATES + 1)) {
	printk("Error copy getsuprates\n");
	islpci_mgt_release(private_config, entry);
	return -EFAULT;
    }

    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_getfixedrate(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *rate = (long *) wrq->u.data.pointer;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_ALOFT_FIXEDRATE, 0, NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_ALOFT_FIXEDRATE,
			    &operation, (void *) &data, &dlen, &entry),
	rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    *rate = (long) *data;
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get Fixed Rate index: %i\n", (int) *rate);
#endif

    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_setfixedrate(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *index = (long *) wrq->u.data.pointer;
    char rate[4] = { 0, 0, 0, 0 };
    void *data;
    unsigned long dlen;
    int rvalue, operation;

    rate[0] = (char) *index;
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set Fixed Rate index: %i\n", (int) rate[0]);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_ALOFT_FIXEDRATE, 0, &rate[0], 4, 0);
    rvalue =
	islpci_mgt_response(private_config, DOT11_OID_ALOFT_FIXEDRATE,
			    &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}


int isl_ioctl_getdtimper(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "->getdtimper\n");
#endif

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET, DOT11_OID_DTIMPERIOD,
		     0, NULL, 4, 0);

    if (rvalue = islpci_mgt_response(private_config, DOT11_OID_DTIMPERIOD,
				     &operation, (void *) &data, &dlen,
				     &entry), rvalue != 0) {
	// an error occurred, return the value
	printk("Error: getdtimper rvalue: %d\n", rvalue);
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	printk("getdtimper: Eoor PIMFOR_OP_ERROR\n");
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    *flag =
	(long) ((*(data + 3) & 0xff) << 24 | (*(data + 2) & 0xff) << 16 |
		(*(data + 1) & 0xff) << 8 | (*data & 0xff));

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get DTIM Period: %lx\n", *flag);
#endif

    islpci_mgt_release(private_config, entry);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "<-getdtimper\n");
#endif

    return 0;
}

int isl_ioctl_setdtimper(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    char period[4] = { 0, 0, 0, 0 };
    void *data;
    unsigned long dlen;
    int rvalue, operation;

    period[0] = (char) (*flag & 0xff);
    period[1] = (char) ((*flag >> 8) & 0xff);
    period[2] = (char) ((*flag >> 16) & 0xff);
    period[3] = (char) ((*flag >> 24) & 0xff);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set DTIM Period: %i\n", (int) period[0]);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET, DOT11_OID_DTIMPERIOD,
		     0, &period[0], 4, 0);
    rvalue = islpci_mgt_response(private_config, DOT11_OID_DTIMPERIOD,
				 &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_getbeaconper(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_BEACONPERIOD, 0, NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_BEACONPERIOD,
			    &operation, (void *) &data, &dlen, &entry),
	rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    *flag = (long) ((*(data + 3) & 0xff) << 24 |
		    (*(data + 2) & 0xff) << 16 |
		    (*(data + 1) & 0xff) << 8 | (*data & 0xff));

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get Beacon Period: %lx\n", *flag);
#endif

    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_setbeaconper(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    char period[4] = { 0, 0, 0, 0 };
    void *data;
    unsigned long dlen;
    int rvalue, operation;

    period[0] = (char) (*flag & 0xff);
    period[1] = (char) ((*flag >> 8) & 0xff);
    period[2] = (char) ((*flag >> 16) & 0xff);
    period[3] = (char) ((*flag >> 24) & 0xff);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set Beacon Period: %i\n", (int) period[0]);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET, DOT11_OID_BEACONPERIOD,
		     0, &period[0], 4, 0);
    rvalue = islpci_mgt_response(private_config, DOT11_OID_BEACONPERIOD,
				 &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_getrate(struct net_device *nwdev, struct iw_param *bitrate)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    double rate;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     GEN_OID_LINKSTATE, 0, NULL, 4, 0);
    if (rvalue = islpci_mgt_response(private_config, GEN_OID_LINKSTATE,
				     &operation, (void *) &data, &dlen,
				     &entry), rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Current rate value %i\n", (int) *data);
#endif

    rate = (double) *data * 500000;
    bitrate->value = (int) rate;

    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_getrtsthresh(struct net_device *nwdev, struct iw_param *rts)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    int rvalue, operation;
    long *data;
    unsigned long dlen;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_RTSTHRESH, 0, NULL, 4, 0);
    if (rvalue = islpci_mgt_response(private_config, DOT11_OID_RTSTHRESH,
				     &operation, (void *) &data, &dlen,
				     &entry), rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    rts->value = *data;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get RTS Thresh 0x%08x (%i)\n", rts->value,
	  rts->value);
#endif

    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_setrtsthresh(struct net_device *nwdev, struct iw_param *rts)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    void *data;
    int rvalue, operation;
    unsigned long dlen;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set RTS Thresh: 0x%08x\n", rts->value);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_RTSTHRESH, 0, &rts->value, 4, 0);
    rvalue = islpci_mgt_response(private_config, DOT11_OID_RTSTHRESH,
				 &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int
isl_ioctl_getfragthresh(struct net_device *nwdev, struct iw_param *frag)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    int rvalue, operation;
    long *data;
    unsigned long dlen;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_FRAGTHRESH, 0, NULL, 4, 0);
    if (rvalue = islpci_mgt_response(private_config, DOT11_OID_FRAGTHRESH,
				     &operation, (void *) &data, &dlen,
				     &entry), rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    frag->value = *data;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get Frag Thresh 0x%08x (%i)\n", frag->value,
	  frag->value);
#endif

    islpci_mgt_release(private_config, entry);
    return 0;
}

int
isl_ioctl_setfragthresh(struct net_device *nwdev, struct iw_param *frag)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    void *data;
    unsigned long dlen;
    int rvalue, operation;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set Frag Thresh: 0x%08x\n", frag->value);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_FRAGTHRESH, 0, &frag->value, 4, 0);
    rvalue = islpci_mgt_response(private_config, DOT11_OID_FRAGTHRESH,
				 &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_wdslinkadd(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    unsigned char *address = (unsigned char *) wrq->u.data.pointer;
    unsigned char macaddr[10];
    void *data;
    unsigned long dlen;
    int rvalue, operation;

    // convert the address string to a MAC address
    string_to_macaddress(address, macaddr);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING,
	  "WDS Link Add addr %02X:%02X:%02X:%02X:%02X:%02X\n", macaddr[0],
	  macaddr[1], macaddr[2], macaddr[3], macaddr[4], macaddr[5]);
#endif

    // transmit the object to the card
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_WDSLINKADD, 0, &macaddr[0], 6, 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    rvalue = islpci_mgt_response(private_config, DOT11_OID_WDSLINKADD,
				 &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_wdslinkdel(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    unsigned char *address = (unsigned char *) wrq->u.data.pointer;
    unsigned char macaddr[10];
    void *data;
    unsigned long dlen;
    int rvalue, operation;

    // convert the address string to a MAC address
    string_to_macaddress(address, macaddr);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING,
	  "WDS Link Remove addr %02X:%02X:%02X:%02X:%02X:%02X\n",
	  macaddr[0], macaddr[1], macaddr[2], macaddr[3], macaddr[4],
	  macaddr[5]);
#endif

    // transmit the object to the card
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_WDSLINKREMOVE, 0, &macaddr[0], 6, 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    rvalue = islpci_mgt_response(private_config, DOT11_OID_WDSLINKREMOVE,
				 &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_eapauthen(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    unsigned char *address = (unsigned char *) wrq->u.data.pointer;
    unsigned char macaddr[10];
    void *data;
    unsigned long dlen;
    int rvalue, operation;

    // convert the address string to a MAC address
    string_to_macaddress(address, macaddr);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING,
	  "EAP Auth STA addr %02X:%02X:%02X:%02X:%02X:%02X\n", macaddr[0],
	  macaddr[1], macaddr[2], macaddr[3], macaddr[4], macaddr[5]);
#endif

    // transmit the object to the card
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_EAPAUTHSTA, 0, &macaddr[0], 6, 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    rvalue = islpci_mgt_response(private_config, DOT11_OID_EAPAUTHSTA,
				 &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_eapunauth(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    unsigned char *address = (unsigned char *) wrq->u.data.pointer;
    unsigned char macaddr[10];
    void *data;
    unsigned long dlen;
    int rvalue, operation;

    // convert the address string to a MAC address
    string_to_macaddress(address, macaddr);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING,
	  "EAP Unauth STA addr %02X:%02X:%02X:%02X:%02X:%02X\n",
	  macaddr[0], macaddr[1], macaddr[2], macaddr[3], macaddr[4],
	  macaddr[5]);
#endif

    // transmit the object to the card
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_EAPUNAUTHSTA, 0, &macaddr[0], 6, 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    rvalue =
	islpci_mgt_response(private_config, DOT11_OID_EAPUNAUTHSTA,
			    &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}


int isl_ioctl_getdot1xen(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_DOT1XENABLE, 0, NULL, 4, 0);
    if (rvalue = islpci_mgt_response(private_config, DOT11_OID_DOT1XENABLE,
				     &operation, (void *) &data, &dlen,
				     &entry), rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get .1x flag: %i\n", (int) *data);
#endif

    *flag = (long) *data;

    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_setdot1xen(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *flag = (long *) wrq->u.data.pointer;
    char enabled[4] = { 0, 0, 0, 0 };
    void *data;
    unsigned long dlen;
    int rvalue, operation;

    enabled[0] = (char) *flag;
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set .1x flag: %i\n", (int) enabled[0]);
#endif

    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_DOT1XENABLE, 0, &enabled[0], 4, 0);
    rvalue = islpci_mgt_response(private_config, DOT11_OID_DOT1XENABLE,
				 &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_getstakeyx(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    unsigned char *parameter = (unsigned char *) wrq->u.data.pointer;
    struct obj_stakey stakey, *pstakey;
    int rvalue, operation;
    unsigned long dlen;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "-> getstakeyx\n");
#endif

    // convert the address string to a MAC address
    string_to_macaddress(parameter, stakey.address);

    // increase the read parameter pointer and read the key number
    parameter += 13;
    stakey.keyid = *parameter - '0';;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "sta addr %02X:%02X:%02X:%02X:%02X:%02X key %i\n",
	  (unsigned char) stakey.address[0],
	  (unsigned char) stakey.address[1],
	  (unsigned char) stakey.address[2],
	  (unsigned char) stakey.address[3],
	  (unsigned char) stakey.address[4],
	  (unsigned char) stakey.address[5], (int) stakey.keyid);
#endif

    // send a get object to get the stations entry
    islpci_mgt_queue(private_config, PIMFOR_OP_GET, DOT11_OID_STAKEY, 0,
		     &stakey, sizeof(struct obj_stakey), 0);
    if (rvalue = islpci_mgt_response(private_config, DOT11_OID_STAKEY,
				     &operation, (void *) &pstakey, &dlen,
				     &entry), rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "key %s length %i\n", pstakey->key,
	  pstakey->length);
#endif

    parameter = (unsigned char *) wrq->u.data.pointer;
    if (copy_to_user((char *) parameter, (void *) &pstakey->key,
		     sizeof(pstakey->key))) {
	printk("Error copy getsuprates\n");
	islpci_mgt_release(private_config, entry);
	return -EFAULT;
    }

    wrq->u.data.length = sizeof(pstakey->key) + 1;
    parameter += sizeof(pstakey->key);
    *parameter = '\0';
    islpci_mgt_release(private_config, entry);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "<- getstakeyx\n");
#endif

    return 0;
}

int isl_ioctl_setstakeyx(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    unsigned char *parameter = (unsigned char *) wrq->u.data.pointer;
    struct obj_stakey stakey;
    void *data;
    int rvalue, operation;
    unsigned long dlen;

    printk("isl_ioctl_setstakeyx\n");
    // convert the address string to a MAC address
    string_to_macaddress(parameter, stakey.address);

    // increase the read parameter pointer and read the key number
    parameter += 13;
    stakey.keyid = *parameter - '0';

    // increase the read parameter pointer to read the key
    parameter += 2;
    stakey.type = DOT11_PRIV_WEP;
    stakey.length =
	strlen(parameter) >
	sizeof(stakey.key) ? sizeof(stakey.key) : strlen(parameter);
    if (copy_from_user(stakey.key, parameter, stakey.length)) {
	return -EFAULT;
    }
//#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "sta addr %02X:%02X:%02X:%02X:%02X:%02X key %i\n",
	  (unsigned char) stakey.address[0],
	  (unsigned char) stakey.address[1],
	  (unsigned char) stakey.address[2],
	  (unsigned char) stakey.address[3],
	  (unsigned char) stakey.address[4],
	  (unsigned char) stakey.address[5], (int) stakey.keyid);
//#endif

    // send a get object to get the stations entry
    islpci_mgt_queue(private_config, PIMFOR_OP_SET, DOT11_OID_STAKEY, 0,
		     &stakey, sizeof(struct obj_stakey), 0);
    rvalue = islpci_mgt_response(private_config, DOT11_OID_STAKEY,
				 &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_getessid(struct net_device *nwdev, struct iw_point *erq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    struct obj_ssid *essid_obj;
    int rvalue, operation;
    unsigned long dlen;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_FUNCTION_CALLS, "->getessid\n");
#endif

    // request the device for the essid
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_SSID, 0, NULL, sizeof(struct obj_ssid), 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    if (rvalue = islpci_mgt_response(private_config, DOT11_OID_SSID,
				     &operation, (void *) &essid_obj,
				     &dlen, &entry), rvalue != 0) {
	// error in getting the response, return the value
	printk("getessid , Error rvalue = %d\n", rvalue);
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	printk("getessid: PIMFOR_OP_ERROR\n");
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "getessid: SSID: %.*s, length=%d\n",
	  essid_obj->length, essid_obj->octets, essid_obj->length);
#endif

    // set ESSID to ON for Wireless Extensions
    erq->flags = 1;
    erq->length = essid_obj->length + 1;

    if (erq->pointer) {
	if (copy_to_user(erq->pointer, essid_obj->octets, erq->length)) {
	    printk("getessid: Error copy getessid\n");
	    islpci_mgt_release(private_config, entry);
	    return -EFAULT;
	}
    } else {
	printk("getessid: Error Pointer == NULL\n");
    }
    islpci_mgt_release(private_config, entry);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_FUNCTION_CALLS, "<-getessid: SSID: %.*s\n",
	  essid_obj->length, essid_obj->octets);
#endif
    return 0;
}

int isl_ioctl_setessid(struct net_device *nwdev, struct iw_point *erq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    struct obj_ssid essid_obj;
    void *data;
    int rvalue, operation;
    unsigned long dlen;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_FUNCTION_CALLS,
	  "%s: ->setessid, erq->pointer= %.*s, erq->flags=%d, erq->length=%d \n",
	  nwdev->name, erq->length, erq->pointer, erq->flags, erq->length);
#endif

    if (erq->flags) {
	if (erq->length > 32)
	    return -E2BIG;

	if (copy_from_user(essid_obj.octets, erq->pointer, erq->length)) {
	    printk("Error setessid copy_from_user\n");
	    return -EFAULT;
	}
	essid_obj.octets[erq->length] = '\0';
	essid_obj.length = erq->length - 1;
    } else {
	// SSID = 'any'
	essid_obj.length = 0;
    }

#if VERBOSE > SHOW_ERROR_MESSAGES
    if (essid_obj.length == 0) {
	DEBUG(SHOW_TRACING, "%s: SSID: 'Any'\n", nwdev->name);
    } else {
	DEBUG(SHOW_TRACING, "%s: SSID: %.*s, length=%d\n", nwdev->name,
	      essid_obj.length, essid_obj.octets, essid_obj.length);
    }
#endif

    // request the device for the essid
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_SSID, 0, (void *) &essid_obj,
		     sizeof(struct obj_ssid), 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    rvalue =
	islpci_mgt_response(private_config, DOT11_OID_SSID, &operation,
			    &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR) {
	printk("Error setessid PIMFOR_OP_ERROR\n");
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_FUNCTION_CALLS, "%s: <-setessid \n", nwdev->name);
#endif

    return rvalue;
}

int isl_ioctl_getbssid(struct net_device *nwdev, char *id)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    // void *data;
    int rvalue, operation;
    unsigned long dlen;
    int ret = 0;
    unsigned char *data;


#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_FUNCTION_CALLS, "%s: ->getbssid \n", nwdev->name);
#endif

    // request the device for the bssid
    islpci_mgt_queue(private_config, PIMFOR_OP_GET, DOT11_OID_BSSID, 0,
		     NULL, 6, 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    if (rvalue = islpci_mgt_response(private_config, DOT11_OID_BSSID,
				     &operation, (void *) &data, &dlen,
				     &entry), rvalue != 0) {
	// error in getting the response, return the value
	islpci_mgt_release(private_config, entry);
	printk("getbssid: Error rvalue= %d\n", rvalue);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	printk("getbssid: PIMFOR_OP_ERROR\n");
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
    // return the data if the destination buffer is defined
    if (id) {
	memcpy(id, data, ETH_ALEN);
	if (ret != 0) {
	    printk("getbssid: Error  copy, id = %d\n", *id);
	    islpci_mgt_release(private_config, entry);
	    return -EFAULT;
	}
    }

    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR) {
	printk("getbssid: Error PIMFOR_OP_ERROR\n");
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "%s: MAC: %02X:%02X:%02X:%02X:%02X:%02X\n",
	  nwdev->name, *data, *(data + 1), *(data + 2), *(data + 3),
	  *(data + 4), *(data + 5));
#endif

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_FUNCTION_CALLS, "%s: <-getbssid \n", nwdev->name);
#endif
    return 0;
}

int isl_ioctl_setbssid(struct net_device *nwdev, char *id)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    char bssid[6];
    void *data;
    unsigned long dlen;
    int rvalue, operation;

    // prepare the structure for the set object
    if (copy_from_user(&bssid[0], id, 6))
	return -EFAULT;

    // send a set object to set the bssid
    islpci_mgt_queue(private_config, PIMFOR_OP_SET, DOT11_OID_BSSID, 0,
		     &bssid[0], 6, 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_BSSID, &operation,
			    &data, &dlen, &entry), rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
    return rvalue;
}

int isl_ioctl_getmode(struct net_device *nwdev, int *mode)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    int bsstype;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

    // send a get object to get the bsstype
    islpci_mgt_queue(private_config, PIMFOR_OP_GET, DOT11_OID_BSSTYPE, 0,
		     NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_BSSTYPE, &operation,
			    (void *) &data, &dlen, &entry), rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
    // returned the bsstype, store it
    bsstype = (int) *data;
    islpci_mgt_release(private_config, entry);

    // send a get object to get the mode
    islpci_mgt_queue(private_config, PIMFOR_OP_GET, OID_INL_MODE, 0, NULL,
		     4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, OID_INL_MODE, &operation,
			    (void *) &data, &dlen, &entry), rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {

	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "getmode: inl_Mode =(%d) bbs_type= %d\n", bsstype,
	  (int) *data);
#endif

    // translate the cards mode to the wireless mode values as good as possible
    switch ((int) *data) {
    case INL_MODE_AP:
	*mode = IW_MODE_MASTER;
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "getmode: IW_MODE_MASTER (%d)\n", (int) *mode);
#endif
	break;

    case INL_MODE_CLIENT:
	if (bsstype == DOT11_BSSTYPE_IBSS) {
	    *mode = IW_MODE_ADHOC;
#if VERBOSE > SHOW_ERROR_MESSAGES
	    DEBUG(SHOW_TRACING, "getmode: IW_MODE_ADHOC (%d)\n",
		  (int) *mode);
#endif
	} else if (bsstype == DOT11_BSSTYPE_INFRA) {
	    *mode = IW_MODE_INFRA;
#if VERBOSE > SHOW_ERROR_MESSAGES
	    DEBUG(SHOW_TRACING, "getmode: IW_MODE_INFRA (%d)\n",
		  (int) *mode);
#endif
	} else {
	    *mode = IW_MODE_AUTO;
#if VERBOSE > SHOW_ERROR_MESSAGES
	    DEBUG(SHOW_TRACING, "getmode: IW_MODE_AUTO (%d)\n",
		  (int) *mode);
#endif
	}
	break;

    default:			// DOT11_BSSTYPE_ANY
	*mode = IW_MODE_AUTO;
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "getmode: IW_MODE_AUTO (%d)\n", (int) *mode);
#endif
	break;
    }
    islpci_mgt_release(private_config, entry);
    return 0;
}


int isl_ioctl_setmode(struct net_device *nwdev, int *mode)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
//    char cardmode[4] = { 0, 0, 0, 0 };
//    char bsstype[4]  = { 0, 0, 0, 0 };
    long cardmode;
    long bsstype;
    void *data;
    unsigned long dlen;
    int rvalue, operation;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_FUNCTION_CALLS, "->isl_ioctl_setmode\n");
#endif

    switch (*mode) {
    case IW_MODE_ADHOC:
	bsstype = DOT11_BSSTYPE_IBSS;
	cardmode = INL_MODE_CLIENT;
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "setmode: IW_MODE_ADHOC(%d) ", *mode);
#endif
	break;

    case IW_MODE_INFRA:
	bsstype = DOT11_BSSTYPE_INFRA;
	cardmode = INL_MODE_CLIENT;
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "setmode: IW_MODE_INFRA(%d) ", *mode);
#endif
	break;

    case IW_MODE_MASTER:
	bsstype = DOT11_BSSTYPE_INFRA;
	cardmode = INL_MODE_AP;
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "setmode: IW_MODE_MASTER(%d) ", *mode);
#endif
	break;

    case IW_MODE_AUTO:
	bsstype = DOT11_BSSTYPE_ANY;
	cardmode = INL_MODE_CLIENT;
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "setmode: IW_MODE_AUTO(%d) ", *mode);
#endif
	break;

    default:
	{
	    printk("%s, EOPNOTSUPP\n", __FUNCTION__);
	    return -EOPNOTSUPP;
	}
    }

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "-> MODE=%ld, BSSTYPE=%ld\n", cardmode, bsstype);
#endif

    // send a set object to set the mode by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET, OID_INL_MODE, 0,
		     &cardmode, sizeof(cardmode), 0);
    rvalue =
	islpci_mgt_response(private_config, OID_INL_MODE, &operation,
			    &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (rvalue)
	return rvalue;
    else if (operation == PIMFOR_OP_ERROR) {
	printk("%s, PIMFOR_OP_ERROR\n", __FUNCTION__);
	return -EINVAL;
    }
    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET, DOT11_OID_BSSTYPE, 0,
		     &bsstype, sizeof(bsstype), 0);
    rvalue =
	islpci_mgt_response(private_config, DOT11_OID_BSSTYPE, &operation,
			    &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR) {
	printk("%s, PIMFOR_OP_ERROR\n", __FUNCTION__);
	return -EINVAL;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_FUNCTION_CALLS, "<-isl_ioctl_setmode\n");
#endif
    return rvalue;
}

int isl_ioctl_getfreq(struct net_device *nwdev, struct iw_freq *frq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    u32 *data, channel;
    unsigned long dlen;
    int rvalue, operation;

    // request the device for the channel
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_FREQUENCY, 0, NULL, 4, 0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_FREQUENCY,
			    &operation, (void *) &data, &dlen, &entry),
	rvalue != 0) {
	// error in getting the response, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    channel = *data;

    // check whether it concern a channel number or a frequency
    // this becomes difficult because the frequency is in kHz
    if (channel < 1000) {
	// interpret the data as a channel
	frq->m = le32_to_cpu(channel);
	frq->e = 0;
    } else {
	// interpret the data as frequency
	frq->m = le32_to_cpu(channel);
	frq->e = 3;
    }

    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_setfreq(struct net_device *nwdev, struct iw_freq *frq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    char channel[4] = { 0, 0, 0, 0 }, *data;
    long mantisse;
    long oid;
    int exponent;
    int rvalue, operation;
    unsigned long dlen;

    // prepare the structure for the set object
    if (frq->m < 1000) {
	// structure value contains a channel indication
	oid = DOT11_OID_CHANNEL;
	channel[0] = (char) frq->m;
	channel[1] = (char) (frq->m >> 8);
    } else {
	// structure contains a frequency indication
	oid = DOT11_OID_FREQUENCY;

	// mantisse is larger than 1000, so make it kHz
	mantisse = frq->m / 1000;
	exponent = frq->e;

	// process the exponent
	while (exponent > 0) {
	    mantisse *= 10;
	    exponent--;
	}

	// convert the frequency to the right byte order
	channel[0] = (char) mantisse;
	channel[1] = (char) (mantisse >> 8);
	channel[2] = (char) (mantisse >> 16);
	channel[3] = (char) (mantisse >> 24);
    }

    // send a set object to set the channel by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET, oid, 0, &channel[0], 4,
		     0);

    // enter sleep mode after supplying oid and local buffers for reception
    // to the management queue receive mechanism
    rvalue =
	islpci_mgt_response(private_config, oid, &operation,
			    (void *) &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_getpsm(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;

    long *psm = (long *) wrq->u.data.pointer;
    int rvalue, operation;

    long *data;
    unsigned long dlen;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET, DOT11_OID_PSM, 0, NULL,
		     4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_PSM, &operation,
			    (void *) &data, &dlen, &entry), rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    *psm = *data;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get PSM 0x%08lx (%li)\n", *psm, *psm);
#endif


    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_setpsm(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *psm = (long *) wrq->u.data.pointer;
    void *data;
    unsigned long dlen;
    int rvalue, operation;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set PSM: 0x%08lx\n", (long) *psm);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_PSM, 0, psm, 4, 0);
    rvalue = islpci_mgt_response(private_config, DOT11_OID_PSM, &operation,
				 &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_getnonerp(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *erp = (long *) wrq->u.data.pointer;
    int rvalue, operation;
    char *data;
    unsigned long dlen;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_NONERPPROTECTION, 0, NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_NONERPPROTECTION,
			    &operation, (void *) &data, &dlen, &entry),
	rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    *erp = (long) *data;
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get Non ERP Protection: %i\n", (int) *erp);
#endif

    islpci_mgt_release(private_config, entry);
    return 0;
}

int isl_ioctl_setnonerp(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    long *nonerp = (long *) wrq->u.data.pointer;
    char erp[4] = { 0, 0, 0, 0 };
    void *data;
    unsigned long dlen;
    int rvalue, operation;

    erp[0] = (char) *nonerp;
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Set Non ERP Protection %i\n", (int) erp[0]);
#endif

    // send a set object to set the bsstype by placing it in the transmit queue
    islpci_mgt_queue(private_config, PIMFOR_OP_SET,
		     DOT11_OID_NONERPPROTECTION, 0, &erp[0], 4, 0);
    rvalue =
	islpci_mgt_response(private_config, DOT11_OID_NONERPPROTECTION,
			    &operation, &data, &dlen, &entry);
    islpci_mgt_release(private_config, entry);
    if (operation == PIMFOR_OP_ERROR)
	return -EINVAL;
    return rvalue;
}

int isl_ioctl_getcountrystring(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    queue_entry *entry = NULL;
    char *cstring = (char *) wrq->u.data.pointer;
    char *rstring;
    int rvalue, operation;
    unsigned long dlen;

    // send a get object to get the filter setting
    islpci_mgt_queue(private_config, PIMFOR_OP_GET,
		     DOT11_OID_COUNTRYSTRING, 0, NULL, 4, 0);
    if (rvalue =
	islpci_mgt_response(private_config, DOT11_OID_COUNTRYSTRING,
			    &operation, (void *) &rstring, &dlen, &entry),
	rvalue != 0) {
	// an error occurred, return the value
	islpci_mgt_release(private_config, entry);
	return rvalue;
    } else if (operation == PIMFOR_OP_ERROR) {
	islpci_mgt_release(private_config, entry);
	return -EINVAL;
    }

    if (copy_to_user((char *) cstring, (void *) rstring, 4)) {
	printk("Error copy getcountrystring\n");
	islpci_mgt_release(private_config, entry);
	return -EFAULT;
    }
#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Get Country String: %02x %02x %02x %02x\n",
	  rstring[0], rstring[1], rstring[2], rstring[3]);
#endif

    wrq->u.data.length = 4;
    islpci_mgt_release(private_config, entry);
    return 0;
}


int isl_ioctl_setpc_debug(struct net_device *nwdev, struct iwreq *wrq)
{
    // long *erp = (long *) wrq->u.data.pointer;
    printk("Set pc_debug from 0x%02x -> to 0x%02x\n", pc_debug,
	   (int) *wrq->u.data.pointer);
    wrq->u.data.length = 1;
    pc_debug = (int) *wrq->u.data.pointer;
    return 0;
}

int isl_ioctl_resetdevice(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING, "Toggled the device reset flag\n");
#endif

    // set the kernel space reset flag
    private_config->resetdevice = 1;

    return 0;
}

int isl_ioctl_checktraps(struct net_device *nwdev, struct iwreq *wrq)
{
    islpci_private *private_config = nwdev->priv;
    char *frame = (char *) wrq->u.data.pointer;
    queue_entry *entry;
    int device_id, operation, flags, length;
    unsigned long oid;

    // check whether the trap queue is empty
    if (islpci_queue_size(private_config->remapped_device_base,
			  &private_config->trap_rx_queue) == 0) {
	// no entries in the trap queue
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "TRAP receive queue empty\n");
#endif
	return -ENOMSG;
    }
    // read a frame from the queue and return it
    islpci_get_queue(private_config->remapped_device_base,
		     &private_config->trap_rx_queue, &entry);

    // decode the PIMFOR frame for data length
    pimfor_decode_header((pimfor_header *) entry->host_address, &operation,
			 &oid, &device_id, &flags, &length);

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_TRACING,
	  "TRAP: oid 0x%lx, device %i, flags 0x%x length %i\n", oid,
	  device_id, flags, length);
#endif

    // copy the frame to the data buffer
    if (copy_to_user
	(frame, (void *) entry->host_address, PIMFOR_HEADER_SIZE + length))
    {
	printk("Error copy checktraps\n");
	// free the entry to the freeq
	islpci_put_queue(private_config->remapped_device_base,
			 &private_config->mgmt_rx_freeq, entry);

	return -EFAULT;
    }
    // free the entry to the freeq
    islpci_put_queue(private_config->remapped_device_base,
		     &private_config->mgmt_rx_freeq, entry);

    return 0;
}


int isl_ioctl(struct net_device *nwdev, struct ifreq *rq, int cmd)
{
    struct iwreq *wrq = (struct iwreq *) rq;
    islpci_private *private_config = nwdev->priv;
    int error = 0;

#if VERBOSE > SHOW_ERROR_MESSAGES
    void *device = private_config->remapped_device_base;

    DEBUG(SHOW_TRACING, "-- isl_ioctl cmd: 0x%x \n", cmd);
    DEBUG(SHOW_QUEUE_INDEXES, "Mgmt Tx Qs: free[%i] shdw[%i]\n",
	  islpci_queue_size(device, &private_config->mgmt_tx_freeq),
	  islpci_queue_size(device, &private_config->mgmt_tx_shadowq));
#endif

    switch (cmd) {
    case SIOCGIWSTATS:
	{
	    struct iw_statistics *stats;
#if VERBOSE > SHOW_ERROR_MESSAGES
	    DEBUG(SHOW_TRACING, "%s: SIOCGIWSTATS(%x)\n", nwdev->name,
		  cmd);
#endif

	    stats = islpci_get_wireless_stats(nwdev);
	    if (stats != NULL) {
		/* Copy statistics to user buffer */
		if (copy_to_user
		    (wrq->u.data.pointer, stats,
		     sizeof(struct iw_statistics)))
		    return -EFAULT;
		if (wrq->u.data.flags != 0)
		    stats->qual.updated = 0;
		return 0;
	    } else
		return -EOPNOTSUPP;

	    break;
	    return -EOPNOTSUPP;
	}

#if WIRELESS_EXT > 8
    case SIOCSIWENCODE:
	{
	    char keybuf[MAX_KEY_SIZE];
#if VERBOSE > SHOW_ERROR_MESSAGES
	    DEBUG(SHOW_TRACING, "%s: SIOCSIWENCODE(%x)\n", nwdev->name,
		  cmd);
#endif
	    if (wrq->u.encoding.pointer) {
		/* We actually have a key to set */
		if (wrq->u.encoding.length > MAX_KEY_SIZE) {
		    error = -E2BIG;
		    break;
		}
		if (copy_from_user(keybuf, wrq->u.encoding.pointer,
				   wrq->u.encoding.length)) {
		    error = -EFAULT;
		    break;
		}
	    } else if (wrq->u.encoding.length != 0) {
		error = -EINVAL;
		break;
	    }
	    error = ioctl_set_encode(nwdev, &(wrq->u.encoding), keybuf);
	}
	break;

    case SIOCGIWENCODE:
	/* only super-user can see encryption key */
	if (!capable(CAP_NET_ADMIN)) {
	    error = -EPERM;
	    break;
	}
	{
	    char keybuf[MAX_KEY_SIZE];
	    error = ioctl_get_encode(nwdev, &(wrq->u.encoding), keybuf);
	    if (wrq->u.encoding.pointer) {
		if (copy_to_user(wrq->u.encoding.pointer, keybuf,
				 wrq->u.encoding.length))
		    error = -EFAULT;
	    }
	}
	break;
#endif				/* WIRELESS_EXT > 8 */

    case SIOCGIWNWID:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWNWID (%x) not supported\n",
	      nwdev->name, cmd);
#endif
	error = -EOPNOTSUPP;
	break;

    case SIOCGIWRETRY:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWRETRY (%x) not supported\n",
	      nwdev->name, cmd);
#endif
	error = -EOPNOTSUPP;
	break;

    case SIOCSIWNICKN:
	printk("%s: Set Nickname: %s \n", nwdev->name,
	       (char *) wrq->u.data.pointer);
	strcpy(nick_name, (char *) wrq->u.data.pointer);
	break;

    case SIOCGIWNICKN:
	strcpy((char *) wrq->u.data.pointer, nick_name);
	wrq->u.data.length = strlen((char *) &wrq->u.data.pointer);
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: Get Nickname: %s, length=%d\n",
	      nwdev->name, (char *) wrq->u.data.pointer,
	      wrq->u.data.length);
#endif
	break;

    case SIOCGIWTXPOW:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWTXPOW (%x) not supported\n",
	      nwdev->name, cmd);
#endif
	error = -EOPNOTSUPP;
	break;

    case SIOCSIWTXPOW:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCSIWTXPOW (%x) not supported\n",
	      nwdev->name, cmd);
#endif
	error = -EOPNOTSUPP;
	break;
    case SIOCGIWPOWER:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWPOWER (%x) not supported\n",
	      nwdev->name, cmd);
#endif
	error = -EOPNOTSUPP;
	break;

    case SIOCGIWSENS:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWSENS (%x) not supported\n",
	      nwdev->name, cmd);
#endif
	error = -EOPNOTSUPP;
	break;

    case SIOCGIWNAME:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWNAME (%x)\n", nwdev->name, cmd);
#endif
	if (private_config->pci_dev_id == PCIDEVICE_ISL3877)
	    strncpy(wrq->u.name, "PRISM Indigo", IFNAMSIZ);
	else if (private_config->pci_dev_id == PCIDEVICE_ISL3890)
	    strncpy(wrq->u.name, "PRISM Duette", IFNAMSIZ);
	else
	    strncpy(wrq->u.name, "Unsupported card type", IFNAMSIZ);
	break;

    case SIOCGIWRATE:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWRATE (%x)\n", nwdev->name, cmd);
#endif
	error = isl_ioctl_getrate(nwdev, &wrq->u.bitrate);
	break;

    case SIOCGIWRTS:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWRTS(%x)\n", nwdev->name, cmd);
#endif
	error = isl_ioctl_getrtsthresh(nwdev, &wrq->u.rts);
	break;

    case SIOCSIWRTS:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCSIWRTS(%x)\n", nwdev->name, cmd);
#endif
	error = isl_ioctl_setrtsthresh(nwdev, &wrq->u.rts);
	break;

    case SIOCGIWFRAG:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWFRAG(%x)\n", nwdev->name, cmd);
#endif
	error = isl_ioctl_getfragthresh(nwdev, &wrq->u.frag);
	break;

    case SIOCSIWFRAG:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCSIWFRAG(%x)\n", nwdev->name, cmd);
#endif
	error = isl_ioctl_setfragthresh(nwdev, &wrq->u.frag);
	break;

    case SIOCGIWRANGE:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWRANGE(%x)\n", nwdev->name, cmd);
#endif
	error = isl_ioctl_getrange(nwdev,
				   (struct iw_range *) wrq->u.data.
				   pointer);
	wrq->u.data.length = sizeof(struct iw_range);
	break;

    case SIOCGIWESSID:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWESSID(%x)\n", nwdev->name, cmd);
#endif
	// get the essid
	error = isl_ioctl_getessid(nwdev, &wrq->u.essid);
	break;

    case SIOCSIWESSID:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCSIWESSID(%x)\n", nwdev->name, cmd);
#endif

	// set the essid to the value supplied
	error = isl_ioctl_setessid(nwdev, &wrq->u.essid);
	break;

    case SIOCGIWAP:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWAP(%x)\n", nwdev->name, cmd);
#endif
	wrq->u.ap_addr.sa_family = ARPHRD_ETHER;
	wrq->u.data.length = ETH_ALEN;
	memset(wrq->u.ap_addr.sa_data, 0, wrq->u.data.length);
	error = isl_ioctl_getbssid(nwdev, wrq->u.ap_addr.sa_data);
	break;

    case SIOCSIWAP:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCSIWAP(%x)\n", nwdev->name, cmd);
#endif
	error = isl_ioctl_setbssid(nwdev, wrq->u.ap_addr.sa_data);
	break;

    case SIOCGIWMODE:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWMODE(%x)\n", nwdev->name, cmd);
#endif
	error = isl_ioctl_getmode(nwdev, &wrq->u.mode);
	wrq->u.data.length = 1;
	break;

    case SIOCSIWMODE:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCSIWMODE(%x) Mode: %d\n", nwdev->name,
	      cmd, wrq->u.mode);
#endif
	// Changing of the Mode during Operation doesn't work very reliable.
	// Card Firmware hangs, kov@artem.de
	return -EOPNOTSUPP;
	error = isl_ioctl_setmode(nwdev, &wrq->u.mode);
	break;

    case SIOCGIWFREQ:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWFREQ(%x)\n", nwdev->name, cmd);
#endif
	error = isl_ioctl_getfreq(nwdev, &wrq->u.freq);
	break;

    case SIOCSIWFREQ:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCSIWFREQ(%x)\n", nwdev->name, cmd);
#endif
	error = isl_ioctl_setfreq(nwdev, &wrq->u.freq);
	break;

    case SIOCGIWPRIV:
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCGIWPRIV(%x)\n", nwdev->name, cmd);
#endif
	if (wrq->u.data.pointer) {
	    if (error = verify_area(VERIFY_WRITE, wrq->u.data.pointer,
				    sizeof(privtab)), error) {
#if VERBOSE > SHOW_ERROR_MESSAGES
		DEBUG(SHOW_TRACING, "verify_area error %i\n", error);
#endif
		break;
	    }

	    wrq->u.data.length = sizeof(privtab) / sizeof(privtab[0]);
	    if (copy_to_user
		(wrq->u.data.pointer, privtab, sizeof(privtab))) {
		printk("Error SIOCGIWPRIV copy\n");
		error = -EFAULT;
	    }
	}
	break;

    case SIOCIWFIRSTPRIV + 0x00:	// setauthenten
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+0 (setauthenten)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setauthenten(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x01:	// getauthenten
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING,
	      "%s: SIOCIWFIRSTPRIV+1 (getauthenten) , length= %d\n",
	      nwdev->name, wrq->u.data.length);
#endif
	error = isl_ioctl_getauthenten(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x02:	// setunencrypted
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+2 (setunencrypt)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setprivfilter(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x03:	// getunencrypted
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+3 (getunencrypt)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getprivfilter(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x04:	// setprivinvoke
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+4 (setprivinvoke)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setprivinvoke(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x05:	// getprivinvoke
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+5 (getprivinvoke)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getprivinvoke(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x06:	// setdefkeyid
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+6 (setdefkeyid)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setdefkeyid(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x07:	// getdefkeyid
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+7 (getdefkeyid)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getdefkeyid(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x08:	// setdefkeyx
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+8 (setdefkeyx)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setdefkeyx(nwdev, wrq, key_current);
	break;

    case SIOCIWFIRSTPRIV + 0x09:	// getdefkeyx
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+9 (getdefkeyx)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getdefkeyx(nwdev, wrq, key_current);
	break;

    case SIOCIWFIRSTPRIV + 0x0A:	// getprivstat
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+10 (getprivstat)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getprivstat(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x0B:	// getsuprates
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+11 (getsuprates)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getsuprates(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x0C:	// setdtimper
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+12 (setdtimper)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setdtimper(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x0D:	// getdtimper
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+13 (getdtimper)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getdtimper(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x0E:	// setbeaconper
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+14 (setbeaconper)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setbeaconper(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x0F:	// getauthenten
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+15 (getbeaconper)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getbeaconper(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x10:	// wdslinkadd
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+16 (wdslinkadd)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_wdslinkadd(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x11:	// wdslinkdel
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+17 (wdslinkdel)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_wdslinkdel(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x12:	// eapauthen
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+18 (eapauthen)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_eapauthen(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x13:	// eapunauth
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+19 (eapunauth)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_eapunauth(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x14:	// setdot1xen
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+20 (setdot1xen)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setdot1xen(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x15:	// getdot1xen
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+21 (getdot1xen)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getdot1xen(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x16:	// setstakeyx
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+22 (setstakeyx)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setstakeyx(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x17:	// getstakeyx
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+23 (getstakeyx)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getstakeyx(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x18:	// setfixedrate
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+24 (setfixedrate)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setfixedrate(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x19:	// getfixedrate
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+25 (getfixedrate)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getfixedrate(nwdev, wrq);
	break;

/*
        case SIOCIWFIRSTPRIV + 0x18:                    // setfragthresh
#if VERBOSE > SHOW_ERROR_MESSAGES
            DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+24 (setfragthresh)\n",
                nwdev->name);
#endif
            error = isl_ioctl_setfragthresh(nwdev, wrq);
            break;

        case SIOCIWFIRSTPRIV + 0x19:                    // getfragthresh
#if VERBOSE > SHOW_ERROR_MESSAGES
            DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+25 (getfragthresh)\n",
                nwdev->name);
#endif
            error = isl_ioctl_getfragthresh(nwdev, wrq);
            break;
*/
    case SIOCIWFIRSTPRIV + 0x1A:	// setpsm
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+26 (setpsm)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setpsm(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x1B:	// getpsm
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+27 (getpsm)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getpsm(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x1C:	// setnonerp
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+28 (setnonerp)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_setnonerp(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x1D:	// getnonerp
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+29 (getnonerp)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getnonerp(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x1E:	// getcountrystr
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+30 (getcountrystr)\n",
	      nwdev->name);
#endif
	error = isl_ioctl_getcountrystring(nwdev, wrq);
	break;

    case SIOCIWFIRSTPRIV + 0x1F:	// checktraps
#if VERBOSE > SHOW_ERROR_MESSAGES
	//DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+31 (checktraps)\n",  nwdev->name);
	DEBUG(SHOW_TRACING, "%s: SIOCIWFIRSTPRIV+31 (pc_debug)\n",
	      nwdev->name);
#endif
	//error = isl_ioctl_checktraps(nwdev, wrq);
	error = isl_ioctl_setpc_debug(nwdev, wrq);
	break;

    default:
	printk("Error cmd 0x%04x not supported!\n", cmd);
	error = -EOPNOTSUPP;
    }

#if VERBOSE > SHOW_ERROR_MESSAGES
    DEBUG(SHOW_QUEUE_INDEXES,
	  "Mgmt Rx Qs: f[%i] s[%i] i[%i] t[%i] p[%i]\n",
	  islpci_queue_size(device, &private_config->mgmt_rx_freeq),
	  islpci_queue_size(device, &private_config->mgmt_rx_shadowq),
	  islpci_queue_size(device, &private_config->ioctl_rx_queue),
	  islpci_queue_size(device, &private_config->trap_rx_queue),
	  islpci_queue_size(device, &private_config->pimfor_rx_queue));
#endif
    if (error != -EOPNOTSUPP && error != 0)
	printk("Error 0x%04x , cmd: 0x%04x\n", error, cmd);
    return error;

}

#else				// no WIRELESS_IOCTLS

int isl_ioctl(struct net_device *nwdev, struct ifreq *rq, int cmd)
{
    printk("-- isl_ioctl cmd: 0x%x disabled\n", cmd);
    int error = 0;
    switch (cmd) {
    default:
	error = -EOPNOTSUPP;
    }
    return error;
}

#endif				// WIRELESS_IOCTLS
